import jinja2
from enum import Enum
import yaml
from pyspark.sql.types import StructType
from typing import Dict


class JinjaVariables(Enum):
    DATABASE = "database"
    TABLE = "table"
    CHECKPOINT = "checkpoint"
    FILENAME_DATE_FORMAT = "filename_date_format"
    PATH_DATE_FORMAT = "path_date_format"
    CONTAINER = "container"


def render_jinja(data: str, replacements: Dict[JinjaVariables, str]):
    if data and isinstance(data, str):
        replace = {k.value: v for (k, v) in replacements.items()}
        skip = False
        for k, v in replace.items():
            if v is None and "{{" + k + "}}" in data.replace(" ", ""):
                skip = True
                break

        if not skip:
            template: jinja2.Template = jinja2.Template(data)
            data = template.render(replace)

    return data


def load_schema(path: str):
    with open(path, "r", encoding="utf-8") as f:
        schema = yaml.safe_load(f)

    schema = StructType.fromJson(schema)

    return schema


def get_ddl(spark_schema: StructType, header: bool = True):
    if header:
        ddl = [f"{f.name} {f.dataType.simpleString()}" for f in spark_schema.fields]
    else:
        ddl = [
            f"_c{i} {f.dataType.simpleString()}"
            for i, f in enumerate(spark_schema.fields)
        ]

    return ddl


def get_html_table(data: dict):
    rows = []
    for k, v in data.items():
        if isinstance(v, dict):
            for ki, vi in v.items():
                row = f"<tr><td>{k}.{ki}</td><td>{vi}</td></tr>"
                rows.append(row)
        else:
            row = f"<tr><td>{k}</td><td>{v}</td></tr>"
            rows.append(row)

    html = "".join(rows)

    html = f"""
  <table>
    <tr><th>Name</th><th>Source</th><th>Destination</th></tr>
    {html}
  </table>
  """

    return html



from pydantic import BaseModel, Field
from typing import Union, List, Any, Dict
from enum import Enum
import fnmatch


_INDEX_WILDCARD = "*"


class StageType(str, Enum):
    landing = "landing"
    raw = "raw"
    base = "base"


class BaseTable(BaseModel):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

    stage: StageType = Field(...)
    database: str = Field(...)
    name: str = Field(...)
    id: Union[str, List[str]] = Field(default=[])


class Table(BaseTable):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

    depends_on: List[str] = Field(default=[])
    table_properties: Dict[str, str] = Field(default=None)


class TableMapping(BaseModel):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

    destination: Table = Field(...)
    source: Union[Dict[str, Table], Table] = Field(...)


class Tables(BaseModel):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self._load_index()

    table_data: dict = Field(...)
    tables_index: Dict[str, Table] = Field(default={})
    table_properties: Dict[str, str] = Field(default=None)

    @classmethod
    def get_index(
        cls,
        stage: Union[StageType, str] = _INDEX_WILDCARD,
        table=_INDEX_WILDCARD,
        database=_INDEX_WILDCARD,
    ):
        return f"{stage.name}.{database}.{table}"

    def _load_index(self):
        for stage in StageType:
            stage_data = self.table_data.get(stage.value)
            if stage_data:
                stage_table_properties = stage_data.get("table_properties", {})

                for database, tables in stage_data.items():
                    if database == "table_properties":
                        continue

                    for table, table_details in tables.items():
                        # flatten the config structure for a table
                        if not table_details:
                            table_details = {}
                        table_properties = table_details.get("table_properties", {})
                        # table_properties = stage_table_properties | table_properties
                        table_properties = {
                            **stage_table_properties,
                            **table_properties,
                        }
                        table_details["name"] = table
                        table_details["database"] = database
                        table_details["stage"] = stage
                        if table_properties:
                            table_details["table_properties"] = table_properties

                        # create a table object
                        table = Table(**table_details)

                        # index the table object
                        index = f"{stage.value}.{database}.{table.name}"
                        self.tables_index[index] = table

    def lookup_table(self, index: str, first_match: bool = True):
        matches = fnmatch.filter(list(self.tables_index.keys()), index)

        if not matches:
            raise Exception(f"index {index} not found in tables_index")

        if first_match:
            matches = matches[0]
            table = self.tables_index[matches]
            return table
        else:
            tables = [self.tables_index[i] for i in matches]
            return tables

    def get_table_mapping(
        self,
        stage: StageType,
        table=_INDEX_WILDCARD,
        database=_INDEX_WILDCARD,
        index: str = None,
    ):
        if not index:
            index = Tables.get_index(stage, table, database)

        destination = self.lookup_table(index=index, first_match=True)
        source = {}

        try:
            for index in destination.depends_on:
                table = self.lookup_table(index=index, first_match=True)
                source[table.name] = table
        except Exception as e:
            raise Exception("Error looking up dependencies for table {}") from e

        if len(list(source.values())) == 1:
            source = list(source.values())[0]

        return TableMapping(source=source, destination=destination)



from ._config import Config
from ._timeslice import Timeslice, TimesliceNow, TimesliceUtcNow
from ._source import Source
from ._deltalake import DeltaLake
from ._table_index import Table, TableMapping, Tables
from ._stage_type import StageType

__all__ = [
    "Config",
    "Timeslice",
    "TimesliceNow",
    "TimesliceUtcNow",
    "Source",
    "DeltaLake",
    "Table",
    "TableMapping",
    "Tables",
    "StageType",
]
